@@ -14,7 +14,6 @@ from typing import (
     Sequence,
     Tuple,
     Type,
-    TypeVar,
     Union,
     cast,
 )
@@ -53,8 +52,6 @@ from .utils import get_params_from_function
 if TYPE_CHECKING:  # pragma: no cover
     from cloup.constraints import BoundConstraintSpec
 
-CallbackReturnType = TypeVar("CallbackReturnType")
-
 
 def get_install_completion_arguments() -> Tuple[click.Parameter, click.Parameter]:
     install_param, show_param = get_completion_inspect_parameters()
@@ -68,7 +65,7 @@ class Typer:
         self,
         *,
         name: Optional[str] = Default(None),
-        cls: Optional[Type[cloup.Command]] = Default(None),
+        cls: Optional[Type[TyperGroup]] = Default(None),
         invoke_without_command: bool = Default(False),
         no_args_is_help: bool = Default(False),
         subcommand_metavar: Optional[str] = Default(None),
@@ -131,7 +128,7 @@ class Typer:
         self,
         name: Optional[str] = Default(None),
         *,
-        cls: Optional[Type[cloup.Command]] = Default(None),
+        cls: Optional[Type[TyperGroup]] = Default(None),
         invoke_without_command: bool = Default(False),
         no_args_is_help: bool = Default(False),
         subcommand_metavar: Optional[str] = Default(None),
@@ -152,10 +149,8 @@ class Typer:
         align_option_groups: Optional[bool] = Default(None),
         show_constraints: Optional[bool] = Default(None),
     ) -> Callable[[CommandFunctionType], CommandFunctionType]:
-        cloup.Command
-
         def decorator(f: CommandFunctionType) -> CommandFunctionType:
-            self.registered_callback = TyperInfo(
+            info = TyperInfo(
                 self,
                 name=name,
                 cls=cls,
@@ -179,6 +174,8 @@ class Typer:
                 align_option_groups=align_option_groups,
                 show_constraints=show_constraints,
             )
+            self.registered_callback = info
+            cast(Any, f)._typer_info = self
             return f
 
         return decorator
@@ -190,7 +187,7 @@ class Typer:
         aliases: Optional[Iterable[str]] = None,
         section: Optional[cloup.Section] = None,
         constraints: Sequence["BoundConstraintSpec"] = (),
-        cls: Optional[Type[cloup.Command]] = None,
+        cls: Optional[Type[TyperCommand]] = None,
         context_settings: Optional[Dict[Any, Any]] = None,
         formatter_settings: Optional[Dict[str, Any]] = None,
         help: Optional[str] = None,
@@ -208,28 +205,28 @@ class Typer:
             cls = TyperCommand
 
         def decorator(f: CommandFunctionType) -> CommandFunctionType:
-            self.registered_commands.append(
-                CommandInfo(
-                    name=name,
-                    aliases=aliases,
-                    section=section,
-                    constraints=constraints,
-                    cls=cls,
-                    context_settings=context_settings,
-                    formatter_settings=formatter_settings,
-                    callback=f,
-                    help=help,
-                    epilog=epilog,
-                    short_help=short_help,
-                    options_metavar=options_metavar,
-                    add_help_option=add_help_option,
-                    no_args_is_help=no_args_is_help,
-                    hidden=hidden,
-                    deprecated=deprecated,
-                    align_option_groups=align_option_groups,
-                    show_constraints=show_constraints,
-                )
+            info = CommandInfo(
+                name=name,
+                aliases=aliases,
+                section=section,
+                constraints=constraints,
+                cls=cls,
+                context_settings=context_settings,
+                formatter_settings=formatter_settings,
+                callback=f,
+                help=help,
+                epilog=epilog,
+                short_help=short_help,
+                options_metavar=options_metavar,
+                add_help_option=add_help_option,
+                no_args_is_help=no_args_is_help,
+                hidden=hidden,
+                deprecated=deprecated,
+                align_option_groups=align_option_groups,
+                show_constraints=show_constraints,
             )
+            self.registered_commands.append(info)
+            cast(Any, f)._command_info = info
             return f
 
         return decorator
@@ -239,7 +236,7 @@ class Typer:
         typer_instance: "Typer",
         *,
         name: Optional[str] = Default(None),
-        cls: Optional[Type[cloup.Command]] = Default(None),
+        cls: Optional[Type[TyperGroup]] = Default(None),
         invoke_without_command: bool = Default(False),
         no_args_is_help: bool = Default(False),
         subcommand_metavar: Optional[str] = Default(None),
@@ -290,92 +287,54 @@ class Typer:
             )
         )
 
-    def _find_command_info(
-        self, callback: Callable[..., CallbackReturnType]
-    ) -> Optional["CommandInfo"]:
-        for reg_command_info in self.registered_commands:
-            if reg_command_info.callback == callback:
-                return reg_command_info
-        for group in self.registered_groups:
-            if group.typer_instance:
-                command_info = group.typer_instance._find_command_info(callback)
-                if command_info:
-                    return command_info
-        return None
-
-    def _find_typer_info(
-        self, callback: Callable[..., CallbackReturnType]
-    ) -> Optional["TyperInfo"]:
-        if self.registered_callback and self.registered_callback.callback == callback:
-            return self.registered_callback
-        for group in self.registered_groups:
-            if group.typer_instance:
-                typer_info = group.typer_instance._find_typer_info(callback)
-                if typer_info:
-                    return typer_info
-        return None
-
-    def _callback_to_click_command(
-        self, callback: Union[TyperCommand, Callable[..., CallbackReturnType]]
-    ) -> Optional[click.Command]:
-        """
-        Returns the `click.Command` object for a given callable, if it is registered under the given `Typer` nstance.
-
-        If the callback is not a registered command, just returns the callback.
-        """
-
-        command_info = self._find_command_info(callback)
-        if command_info:
-            return get_command_from_info(command_info)
-        else:
-            typer_info = self._find_typer_info(callback)
-            if typer_info:
-                typer_info.typer_instance = self
-                return get_group_from_info(typer_info)
-        return None
-
     def invoke(
         self,
-        callback: Union[TyperCommand, Callable[..., CallbackReturnType]],
+        callback: Callable[..., Any],
         *args: Any,
         **kwargs: Any,
-    ) -> CallbackReturnType:
+    ) -> Any:
         """
         Invokes a callable that is a registered command or subcommand of the given `Typer` instance.
 
         Other arguments are passed on to `click.Context.invoke`.
         """
 
-        command = self._callback_to_click_command(callback) or callback
+        info = getattr(callback, "_command_info", None)
+        if not info:
+            raise TypeError("callback is not a command")
+
+        command = get_command_from_info(info)
 
         ctx = click.get_current_context()
-        return cast(CallbackReturnType, ctx.invoke(command, *args, **kwargs))
+        return ctx.invoke(command, *args, **kwargs)
 
     def forward(
         self,
-        callback: Union[TyperCommand, Callable[..., CallbackReturnType]],
+        callback: Callable[..., Any],
         *args: Any,
         **kwargs: Any,
-    ) -> CallbackReturnType:
+    ) -> Any:
         """
         Forwards a callable that is a registered command or subcommand of the given `Typer` instance.
 
         Other arguments are passed on to `click.Context.forward`.
         """
 
-        command = self._callback_to_click_command(callback)
-        if not command:
+        info = getattr(callback, "_command_info", None)
+        if not info:
             raise TypeError("callback is not a command")
 
+        command = get_command_from_info(info)
+
         ctx = click.get_current_context()
-        return cast(CallbackReturnType, ctx.forward(command, *args, **kwargs))
+        return ctx.forward(command, *args, **kwargs)
 
     def __call__(self, *args: Any, **kwargs: Any) -> Any:
         return get_command(self)(*args, **kwargs)
 
 
 def process_help_text(command: click.Command) -> None:
-    if not docstring_parser:
+    if not docstring_parser:  # pragma: no cover
         return None
 
     if command.help is None:
@@ -568,7 +527,6 @@ def get_group_from_info(group_info: TyperInfo) -> click.Group:
             align_sections=solved_info.align_sections,
             show_subcommand_aliases=solved_info.show_subcommand_aliases,
             constraints=get_constraints(
-                cls=cls,
                 callback=solved_info.callback,
                 constraints=solved_info.constraints,
             ),
@@ -637,7 +595,7 @@ def get_params_convertors_ctx_param_name_from_function(
                 for param_name in opt_group_params:
                     option = params.get(param_name, None)
                     if option is None:
-                        raise ValueError("{param} is not a parameter")
+                        raise ValueError(f"`{param_name}` is not a parameter")
                     if not isinstance(option, click.Option):
                         raise TypeError(
                             "only parameter of type `Option` can be added to option groups"
@@ -672,7 +630,6 @@ def get_command_from_info(command_info: CommandInfo) -> click.Command:
     if issubclass(cls, cloup.Command):
         kwargs = dict(
             constraints=get_constraints(
-                cls=cls,
                 callback=command_info.callback,
                 constraints=command_info.constraints,
             ),
@@ -790,7 +747,6 @@ def get_callback(
 
 def get_constraints(
     *,
-    cls: Optional[Type[cloup.Command]] = None,
     callback: Optional[Callable[..., Any]] = None,
     constraints: Sequence["BoundConstraintSpec"] = (),
 ) -> Sequence["BoundConstraintSpec"]:
@@ -798,11 +754,6 @@ def get_constraints(
         return constraints
     callback_obj = cast(Any, callback)
     if hasattr(callback_obj, "__constraints"):
-        if cls and not issubclass(cls, cloup.ConstraintMixin):
-            raise TypeError(
-                f"a Command must inherit from `cloup.ConstraintMixin` to support "
-                f"constraints; {cls} doesn't"
-            )
         decorator_constraints = tuple(reversed(callback_obj.__constraints))
         return tuple(constraints) + decorator_constraints
 
@@ -942,7 +893,7 @@ def get_cloup_param(
                 if type_ is NoneType:
                     continue
                 types.append(type_)
-            assert len(types) == 1, "Typer Currently doesn't support Union types"
+            assert len(types) == 1, "Typer currently doesn't support `Union` types"
             main_type = types[0]
             origin = getattr(main_type, "__origin__", None)
         # Handle Tuples and Lists
@@ -1021,7 +972,7 @@ def get_cloup_param(
                 expose_value=parameter_info.expose_value,
                 is_eager=parameter_info.is_eager,
                 envvar=parameter_info.envvar,
-                shell_complete=parameter_info.shell_complete,
+                shell_complete=get_param_completion(parameter_info.shell_complete),
             ),
             convertor,
         )
@@ -1052,6 +1003,7 @@ def get_cloup_param(
                 expose_value=parameter_info.expose_value,
                 is_eager=parameter_info.is_eager,
                 envvar=parameter_info.envvar,
+                shell_complete=get_param_completion(parameter_info.shell_complete),
             ),
             convertor,
         )
@@ -1117,16 +1069,15 @@ def get_param_completion(
         return None
     parameters = get_params_from_function(callback)
     ctx_name = None
-    args_name = None
+    param_name = None
     incomplete_name = None
     unassigned_params = [param for param in parameters.values()]
     for param_sig in unassigned_params[:]:
-        origin = getattr(param_sig.annotation, "__origin__", None)
         if lenient_issubclass(param_sig.annotation, click.Context):
             ctx_name = param_sig.name
             unassigned_params.remove(param_sig)
-        elif lenient_issubclass(origin, List):
-            args_name = param_sig.name
+        elif lenient_issubclass(param_sig.annotation, click.Parameter):
+            param_name = param_sig.name
             unassigned_params.remove(param_sig)
         elif lenient_issubclass(param_sig.annotation, str):
             incomplete_name = param_sig.name
@@ -1136,8 +1087,8 @@ def get_param_completion(
         if ctx_name is None and param_sig.name == "ctx":
             ctx_name = param_sig.name
             unassigned_params.remove(param_sig)
-        elif args_name is None and param_sig.name == "args":
-            args_name = param_sig.name
+        elif param_name is None and param_sig.name == "param":
+            param_name = param_sig.name
             unassigned_params.remove(param_sig)
         elif incomplete_name is None and param_sig.name == "incomplete":
             incomplete_name = param_sig.name
@@ -1149,12 +1100,14 @@ def get_param_completion(
             f"Invalid shell-completion callback parameters: {show_params}"
         )
 
-    def wrapper(ctx: click.Context, args: List[str], incomplete: Optional[str]) -> Any:
+    def wrapper(
+        ctx: click.Context, param: click.Parameter, incomplete: Optional[str]
+    ) -> Any:
         use_params: Dict[str, Any] = {}
         if ctx_name:
             use_params[ctx_name] = ctx
-        if args_name:
-            use_params[args_name] = args
+        if param_name:
+            use_params[param_name] = param
         if incomplete_name:
             use_params[incomplete_name] = incomplete
         return callback(**use_params)  # type: ignore
@@ -1164,6 +1117,6 @@ def get_param_completion(
 
 
 def run(function: Callable[..., Any]) -> None:
-    app = Typer()
+    app = Typer(add_completion=False)
     app.command()(function)
     app()
